/* (C) Copyright 1984,85,86,87 Walter L. Peacock   All Rights Reserved   */
/*  B L B 2 K E Y S . C   10/30/87  */

/*************************

If this program is invoked with a "-v" option it will operate in
variable-length record mode.

The following three files must be in the current directory in order to
run this program:

   o  BTPARMS.BTR    -  CBTREE parameter file,
   o  PATIENT.RAW    -  Raw data file,
   o  CKERROR.MSG    -  error message file.

This program creates a fixed or variable-length record data file
(PATIENT.DAT or PATIENT.VAR) and creates two B+tree indexes on the specified
data file.

Instead of requiring operator input, the raw data is fetched from the file
PATIENT.RAW.  The working data file PATIENT.DAT/PATIENT.VAR is created in
coordination with the insertion of keys into the B+trees.

Before running your application, the BTSETUP.EXE utility must be executed to
specify the B+tree parameters for our index files.  You should specify a
data file record length of 0 (zero) to activate variable length record
processing.  If you desire to place both B+trees in a single index file,
then the Index File Name should be the same for both B+trees, as should also
be the Index Block Size.

The parameters specified with BTSETUP.EXE are saved in the BTPARMS.BTR file.
The function btrinit(), called below, will read the specified entry in this
file to obtain the proper B+tree parameters.  These parameters are then used
to initialize the B+tree communication structure (struct btcommo) which will
be used for the given B+tree.

General summary of CBTREE application usage:

   1. Include "cbtree.h" in all applications using CBTREE.
   2. Declare a btcommo structure (BTC) for each B+tree.
   3. Declare file descriptors for each file (index and data).
   4. Call the function btrinit() to initialize the btcommo structure.
   5. To create a new, initialized, B+tree index file, call the function
      creatbtr() immediately after calling btrinit().
   6. Use bt_open() (see "btfio.h" and "btopen.c") to open your index and
      data files.
   7. Use initdat() to initialize the data file header: If the INSERT or
      DELETE operations are to be used, the first record of your data file
      is reserved for use by CBTREE.  The function initdat() will initialize
      this header record to it's proper value.

      For each CBTREE operation:

   8. Specify the CBTREE operation type in btcommo.btoptype.
   9. Construct your key into btcommo.btkey.
  10. Set btcommo.btloc according to the operation (see manual).
  11. If your data file has variable length records and you are making an
      INSERT operation, btcommo.btdtalen must be set to the length of the
      corresponding data record.
  12. Issue your call to cbtree().
  13. For retrieval calls, the record location corresponding to the
      specified key is returned in btcommo.btloc (or btcommo.btrecnum[] if a
      block retrieval operation, see manual).
  14. Use the macro btseek() or (btseekv() for variable length records)
      to position yourself at the specified data record.
  15. Before reading a variable length record, use the function btsizvar()
      to determine the size of the record at the given location.
  16. Repeat steps 8 through 15, as needed.
  17. Call btrterm() to release resources used by the btcommo structure.
  18. Call close() to close files and exit.

**************************/

/* Include standard header files */

#include <stdio.h>

/* Include special CBTREE header files */

#include "cbtree.h"

/* btfio.h *must* be included when bt_open() is used */

#include "btfio.h"

/* Include other application-specific header files */

#include "patrec.str"   /* patient data record definition */

#if CI | DC
#else
#include <ctype.h>
#endif

#if AmigaDOS & LC
#define printf iprintf  /* use fast integer printf routine */
#endif

/******* start here *******/

void main(argc, argv)
int   argc;
char *argv[];
{
#if LOHI
#else
   extern long  fliplong();   /* for high-byte/low-byte machines */
#endif
   extern long  atol();
   extern int   bt_open(), close();
   extern uword patvar();
   extern char *calloc(), *strnncpy();
   extern void  free();
   struct btcommo btnme, btssn;  /* A btcommo structure (BTC) is  */
         /* required for each B+tree index your applicatoin will be  */
         /* using.  Call btrinit() for each btcommo structure (BTC). */
   PATREC   pr;            /* patient record structure: see patrec.str  */
   BTBLKHDR freespc;       /* The BTBLKHDR structure 'freespc' (see     */
         /* cbtree.h) is used to process the header records for both    */
         /* data and index files.  It is used in conjunction with the   */
         /* the function putfhdr() which will initialize you header     */
         /* record.  See example code below. */
   long     rn, rcnt, saveof;
            /* file DESCRIPTORS are required for use by all CBTREE files. */
   int      fdraw, fddat, fdnme, fdssn;
   int      btretcd, varlen;
   char    *nam_btdat, *nam_btnme, *nam_btssn;
   char     ans[14];
   char     prv[sizeof(PATREC)];    /* buffer as big as a patient record */
               /* to hold compressed (variable-length) version of record */

   varlen = (argc > 1 && strcmp(argv[1], "-v") == 0);

   if (varlen)
   {
      nam_btdat = "patient.var"; /* data file name */
      nam_btnme = "btvarnme";    /* NAME B+tree name */
      nam_btssn = "btvarssn";    /* SSN  B+tree name */
   }
   else
   {
      nam_btdat = "patient.dat";
      nam_btnme = "btpatnme";
      nam_btssn = "btpatssn";
   }

   wopen("CON:0/0/640/170/Bld2keys");     /* open Amiga window */

   scr_clr();   /* clear the screen */
   scr_curs(2,29);
   printf("BLD2KEYS");
   scr_curs(4, 10);
   printf("o   Initializing btcommo structure for NAME B+tree...");

   /* Call btrinit() for each BTC (btcommo structure) defined above. */
   /* The first parameter is the B+tree Name as it was defined by the   */
   /* BTSETUP.EXE support utility.  The second parameter is the address */
   /* of the BTC (btcommo structure) defined above.  Be sure to ALWAYS  */
   /* check the return call to btrinit() ! */

   if( (btrinit(nam_btnme, &btnme) ) == ERR)
   {
      ckerror( CK_BTRIN, "bld2var: btnme");
      goto b2v_0exit;
   }

   scr_curs(5, 10);
   printf("o   Creating NAME B+tree in '%s' file...", btnme.idxname);

   /* Call creatbtr() to initialize a new index file or to add a B+tree */
   /* to an existing file.  Creatbtr() should be call immediately */
   /* following it's corresponding btrinit() call.  When processing an  */
   /* existing B+tree, skip creatbtr().   */

   creatbtr(&btnme);   /* creatbtr() must immediately follow btrinit() */

   scr_curs(6, 10);
   printf("o   Initializing btcommo structure for SSN B+tree...");

   /* Second btrinit() for second B+tree */

   if( (btrinit(nam_btssn, &btssn) ) == ERR)
   {
      ckerror( CK_BTRIN, "bld2var: btssn");
      goto b2v_1exit;
   }

   scr_curs(7, 10);
   printf("o   Creating SSN B+tree in '%s' file...", btssn.idxname);

   /* And second creatbtr() */

   creatbtr(&btssn);

   /* Use bt_open() to open all your index and data files.  This will   */
   /* open your files in a portable manner and ensure that the  */
   /* 'untranslated' mode is set.  You would normally use the mode   */
   /* flags:   */
   /*          */
   /*    These constants are defined in btfio.h:   */
   /*          */
   /*       O_RDONLY    open file for reading only */
   /*       O_RDWR      open file for reading and writing   */
   /*       O_NEW       Create/truncate new file   */
   /*          */
   /* With O_NEW you *must* specify the permission flag S_IRDWR as a    */
   /* third parameter to bt_open() */

   if( (fdraw = bt_open("patient.raw",O_RDONLY) ) == ERR)   /* Read Only */
   {
      ckerror( CK_OPEN, "bld2var: PATIENT.RAW");
      goto b2v_2exit;
   }

   /* Create a new "patient.var" file for reading and writing, with */
   /* file access permission set for both read access and write access. */

   if( (fddat = bt_open(nam_btdat, O_NEW|O_RDWR, S_IRDWR) ) == ERR)
   {
      ckerror( CK_OPEN, "bld2var: nam_btdat");
      goto b2v_3exit;
   }

   /* The index file name was specified when you ran BTSETUP.EXE.  This */
   /* name was stored in BTPARMS.BTR.  It was then, in turn, placed in  */
   /* your btcommo structure by the function btrinit() (called above).  */
   /* When opening your index file, you should specify the name given   */
   /* by btcommo.idxname as follows:  */

   /* This file was created by creatbtr() (above) so we do not specify  */
   /* O_NEW in the bt_open() call.  */

   if( (fdnme = bt_open(btnme.idxname, O_RDWR) ) == ERR)
   {
      ckerror( CK_OPEN, "bld2var: btnme.idxname");
      goto b2v_4exit;
   }

   /* Now open the secondary index file in the same way: */

   if (strcmp(btssn.idxname, btnme.idxname) == 0)  /* same file name ? */
      fdssn = fdnme;    /* same file */
   else if( (fdssn = bt_open(btssn.idxname, O_RDWR) ) == ERR)
   {
      ckerror( CK_OPEN, "bld2var: btssn.idxname");
      goto b2v_5exit;
   }

   /* Initialize the datafile header record. */

   initdat(fddat, &btnme);   /* initialize the header record */

   /* Read header pointers of patient.raw to determine last record */

   getfhdr(&freespc, fdraw);      /* get patient.raw end of file marker */
#if LOHI
   saveof = freespc.eoflst - 1;   /* last record to process */
#else
   saveof = fliplong(freespc.eoflst) - 1;   /* For 680x0 machines, etc. */
#endif

   /* Ask how many records to copy from patient.raw */

   scr_curs(9, 14);
   printf("Number of records to copy ==> ");
   gets(ans);
   if ((rcnt = atol(ans)) == 0L)
      goto b2v_6exit;

   /* tell viewer what's going on */

   scr_curs(11, 10);
   printf("o   Reading data record from PATIENT.RAW,");
   scr_curs(12, 10);
   printf("o   INSERT-ing NAME key into '%s',", btnme.idxname);
   scr_curs(13, 10);
   printf("o   Writing data record to '%s',", nam_btdat);
   scr_curs(14, 10);
   printf("o   ISRTKY-ing SSN key into '%s'.", btssn.idxname);

   /* Set the operation type for each of the B+trees. */

   /* The primary key uses INSERT which inserts the key into the B+tree */
   /* but also asks CBTREE to specify what the next available DATA   */
   /* record location is. This is done automatically during the INSERT  */
   /* operation by referring to the datafile free-list and end-of-file  */
   /* pointers.  The INSERT operation tells you where to write your  */
   /* data record as the return value in btcommo.btloc.   */
   /* You must write your data record AFTER calling INSERT. */

   /* When using variable length records, the record length should be   */
   /* specified in btcommo.btdtalen before calling cbtree() (see below). */

   btnme.btoptype = INSERT;      /* INSERT for primary keys (btnme)  */

   /* The secondary B+tree uses the ISRTKY operation to insert the key  */
   /* only without modifying the datafile.   */

   /* ISRTKY must be told (in btcommo.btloc) where you are storing your */
   /* data record.  This information was typically just given to you by */
   /* the INSERT operation.  See below for details. */

   btssn.btoptype = ISRTKY;      /* ISRTKY for secondary keys (btssn) */

   rn = 1;  /* loop/record counter */
   while (rcnt)
   {
      if (rn >= saveof) /* wrap around if at the end of patient.raw */
         rn = 1;        /* and reuse the same records. */

      scr_curs(11, 10);          /* tell em what we're doing */
      getfpat(&pr, fdraw, ++rn); /* read the record from patient.raw */

      /* Now build the key from the raw (fixed-length) record. */
      /* Bldkey.c contains an example of how to build a concatenated */
      /* key from several fields in a record.  Note the necessity of */
      /* using blanks between fields if variable length fields are   */
      /* used.  Key-length is a run-time parameter.  This means that */
      /* You can change the key-length expected by your application  */
      /* without recompiling.  We recommend that you write your key- */
      /* building function similarly to bldkey.c.  */

      bldkey(btnme.btkey, btnme.btkeylen, &pr); /* key */

      /* We will create a variable length record out of the fixed-   */
      /* length version to illustrate the use of variable length     */
      /* records.    */

      if (varlen)    /* set record length BEFORE calling INSERT ! */
         btnme.btdtalen = patvar(prv, &pr);  /* variable length record */

      /* issue cbtree INSERT call */

      scr_curs(12, 10);
      if ((btretcd = cbtree(fddat, fdnme, &btnme)) != BTCALLOK)
         cberror(btretcd, "bld2var", &btnme);
      else
      {
         /* Now we know WHERE to write our data record. */
         /* putfpatv() calls btseekv() which uses information in  */
         /* the btcommo structure to correctly position us at the */
         /* specified record location */

         scr_curs(13, 10);
         if (varlen)
            putfpatv(fddat, &btnme, prv); /* put variable length record */
         else
            putfpat(fddat, btnme.btloc, &pr);   /* fixed length */

         /* Now build the second key using a simple strnncpy()  */
         /* Make sure your key is NUL-terminated!  Not all  */
         /* versions of strncpy() guarantee this.  The CBTREE library */
         /* contains one that does. */

         strnncpy(btssn.btkey, pr.ssn, btssn.btkeylen);

         /* Set the btloc value in preparation for the ISRTKY call.  */
         /* Notice that it is the same value returned by the INSERT  */
         /* operation, but must be copied from a different btcommo   */
         /* structure!  */

         btssn.btloc = btnme.btloc;    /* tell location to other tree */

         /* issue cbtree ISTRKY call */

         scr_curs(14, 10);
         if ((btretcd = cbtree(fddat, fdssn, &btssn)) != BTCALLOK)
            cberror(btretcd, "bld2var", &btssn);

         scr_curs(16, 27);
         printf("%5ld", rcnt--);
#if MW
         fflush(stdout);
#endif
      }
   }

b2v_6exit:
   close(fdssn);        /* close files, and... */
b2v_5exit:
   if (fdnme != fdssn)
      close(fdnme);
b2v_4exit:
   close(fddat);
b2v_3exit:
   close(fdraw);
b2v_2exit:
   btrterm(&btnme);     /* release resources */
b2v_1exit:
   btrterm(&btssn);
b2v_0exit:

   wclose();            /* close Amiga window */
   puts("");
}


/* The file PATIENT.RAW is distributed in 808x form which has low-byte */
/* high-byte order.  This must be reversed for 680x0 machines.         */
#if LOHI
#else

#define SZLNG (sizeof(long))
union onelong
{
   long lng;
   char chr[SZLNG];
};

long fliplong(l)
long l;
{
   union onelong l1, l2;
   int i;

   l1.lng = l;                         /* get the long */
   for (i = SZLNG; i-- != 0; )
      l2.chr[SZLNG-i-1] = l1.chr[i];   /* flip it */
   return (l2.lng);
}

#endif

